@@ -5,27 +5,40 @@ import scipy.sparse
 import tensorflow as tf
 
 from sklearn.model_selection import train_test_split
-from typing import Optional, Dict, Text, List, Tuple, Any, Union, Generator, NamedTuple
+from typing import (
+    Optional,
+    Dict,
+    Text,
+    List,
+    Tuple,
+    Any,
+    Union,
+    Generator,
+    NamedTuple,
+    ValuesView,
+    ItemsView,
+)
 from collections import defaultdict
 from rasa.utils.tensorflow.constants import BALANCED, SEQUENCE
 
 logger = logging.getLogger(__name__)
 
 
-# Mapping of feature name to a list of numpy arrays representing the actual features
+# Mapping of attribute name and feature name to a list of numpy arrays representing
+# the actual features
 # For example:
-# "text_features" -> [
+# "text" -> { "sentence": [
 #   "numpy array containing dense features for every training example",
 #   "numpy array containing sparse features for every training example"
-# ]
-Data = Dict[Text, List[np.ndarray]]
+# ]}
+Data = Dict[Text, Dict[Text, List[np.ndarray]]]
 
 
 class FeatureSignature(NamedTuple):
     """Stores the shape and the type (sparse vs dense) of features."""
 
     is_sparse: bool
-    shape: List[int]
+    feature_dimension: Optional[int]
 
 
 class RasaModelData:
@@ -35,50 +48,109 @@ class RasaModelData:
     """
 
     def __init__(
-        self, label_key: Optional[Text] = None, data: Optional[Data] = None
+        self,
+        label_key: Optional[Text] = None,
+        label_sub_key: Optional[Text] = None,
+        data: Optional[Data] = None,
     ) -> None:
         """
         Initializes the RasaModelData object.
 
         Args:
-            label_key: the label_key used for balancing, etc.
+            label_key: the key of a label used for balancing, etc.
+            label_sub_key: the sub key of a label used for balancing, etc.
             data: the data holding the features
         """
 
-        self.data = data or {}
+        self.data = data or defaultdict(lambda: defaultdict(list))
         self.label_key = label_key
+        self.label_sub_key = label_sub_key
         # should be updated when features are added
         self.num_examples = self.number_of_examples()
 
-    def get_only(self, key: Text) -> Optional[np.ndarray]:
-        if key in self.data:
-            return self.data[key][0]
-        else:
-            return None
+    def get(
+        self, key: Text, sub_key: Optional[Text] = None
+    ) -> Union[Dict[Text, List[np.ndarray]], List[np.ndarray]]:
+        """Get the data under the given keys.
 
-    def get(self, key: Text) -> List[np.ndarray]:
-        if key in self.data:
+        Args:
+            key: The key.
+            sub_key: The optional sub key.
+
+        Returns:
+            The requested data.
+        """
+        if sub_key is None and key in self.data:
             return self.data[key]
-        else:
-            return []
 
-    def items(self):
+        if sub_key and key in self.data and sub_key in self.data[key]:
+            return self.data[key][sub_key]
+
+        return []
+
+    def items(self) -> ItemsView:
+        """Return the items of the data attribute.
+
+        Returns:
+            The items of data.
+        """
         return self.data.items()
 
-    def values(self):
+    def values(self) -> ValuesView[Dict[Text, List[np.ndarray]]]:
+        """Return the values of the data attribute.
+
+        Returns:
+            The values of data.
+        """
         return self.data.values()
 
-    def keys(self):
-        return self.data.keys()
+    def keys(self, key: Optional[Text] = None) -> List[Text]:
+        """Return the keys of the data attribute.
+
+        Args:
+            key: The optional key.
+
+        Returns:
+            The keys of the data.
+        """
+        if key is None:
+            return list(self.data.keys())
+
+        if key in self.data:
+            return list(self.data[key].keys())
+
+        return []
 
     def first_data_example(self) -> Data:
-        return {
-            feature_name: [feature[:1] for feature in features]
-            for feature_name, features in self.data.items()
-        }
+        """Return the data with just one feature example per key, sub-key.
 
-    def feature_not_exist(self, key: Text) -> bool:
-        """Check if feature key is present and features are available."""
+        Returns:
+            The simplified data.
+        """
+        out_data = {}
+        for key, attribute_data in self.data.items():
+            out_data[key] = {}
+            for sub_key, features in attribute_data.items():
+                out_data[key][sub_key] = [feature[:1] for feature in features]
+        return out_data
+
+    def does_feature_not_exist(self, key: Text, sub_key: Optional[Text] = None) -> bool:
+        """Check if feature key (and sub-key) is present and features are available.
+
+        Args:
+            key: The key.
+            sub_key: The optional sub-key.
+
+        Returns:
+            True, if no features for the given keys exists, False otherwise.
+        """
+        if sub_key:
+            return (
+                key not in self.data
+                or not self.data[key]
+                or sub_key not in self.data[key]
+                or not self.data[key][sub_key]
+            )
 
         return key not in self.data or not self.data[key]
 
@@ -90,16 +162,29 @@ class RasaModelData:
     def number_of_examples(self, data: Optional[Data] = None) -> int:
         """Obtain number of examples in data.
 
-        Raises: A ValueError if number of examples differ for different features.
-        """
+        Args:
+            data: The data.
 
+        Raises: A ValueError if number of examples differ for different features.
+
+        Returns:
+            The number of examples in data.
+        """
         if not data:
             data = self.data
 
         if not data:
             return 0
 
-        example_lengths = [v.shape[0] for values in data.values() for v in values]
+        example_lengths = [
+            f.shape[0]
+            for attribute_data in data.values()
+            for features in attribute_data.values()
+            for f in features
+        ]
+
+        if not example_lengths:
+            return 0
 
         # check if number of examples is the same for all values
         if not all(length == example_lengths[0] for length in example_lengths):
@@ -110,68 +195,124 @@ class RasaModelData:
 
         return example_lengths[0]
 
-    def feature_dimension(self, key: Text) -> int:
-        """Get the feature dimension of the given key."""
+    def feature_dimension(self, key: Text, sub_key: Text) -> int:
+        """Get the feature dimension of the given key.
+
+        Args:
+            key: The key.
+            sub_key: The optional sub-key.
+
+        Returns:
+            The feature dimension.
+        """
+        if key not in self.data or sub_key not in self.data[key]:
+            return 0
 
         number_of_features = 0
-        for data in self.data[key]:
+        for data in self.data[key][sub_key]:
             if data.size > 0:
                 number_of_features += data[0].shape[-1]
 
         return number_of_features
 
-    def add_features(self, key: Text, features: List[np.ndarray]):
+    def add_data(self, data: Data, key_prefix: Optional[Text] = None) -> None:
+        """Add incoming data to data.
+
+        Args:
+            data: The data to add.
+            key_prefix: Optional key prefix to use in front of the key value.
+        """
+        for key, attribute_data in data.items():
+            for sub_key, features in attribute_data.items():
+                if key_prefix:
+                    self.add_features(f"{key_prefix}{key}", sub_key, features)
+                else:
+                    self.add_features(key, sub_key, features)
+
+    def add_features(
+        self, key: Text, sub_key: Text, features: Optional[List[np.ndarray]]
+    ) -> None:
         """Add list of features to data under specified key.
 
         Should update number of examples.
+
+        Args:
+            key: The key
+            sub_key: The sub-key
+            features: The features to add.
         """
-
-        if not features:
+        if features is None:
             return
 
-        if key in self.data:
-            raise ValueError(f"Key '{key}' already exists in RasaModelData.")
-
-        self.data[key] = []
-
         for data in features:
             if data.size > 0:
-                self.data[key].append(data)
+                self.data[key][sub_key].append(data)
 
-        if not self.data[key]:
-            del self.data[key]
+        if not self.data[key][sub_key]:
+            del self.data[key][sub_key]
 
         # update number of examples
         self.num_examples = self.number_of_examples()
 
-    def add_lengths(self, key: Text, from_key: Text) -> None:
-        """Adds np.array of lengths of sequences to data under given key."""
-        if not self.data.get(from_key):
+    def add_lengths(
+        self, key: Text, sub_key: Text, from_key: Text, from_sub_key: Text
+    ) -> None:
+        """Adds np.array of lengths of sequences to data under given key.
+
+        Args:
+            key: The key to add the lengths to
+            sub_key: The sub-key to add the lengths to
+            from_key: The key to take the lengths from
+            from_sub_key: The sub-key to take the lengths from
+        """
+        if not self.data.get(from_key) or not self.data.get(from_key, {}).get(
+            from_sub_key
+        ):
             return
 
-        self.data[key] = []
+        self.data[key][sub_key] = []
 
-        for data in self.data[from_key]:
+        for data in self.data[from_key][from_sub_key]:
             if data.size > 0:
                 lengths = np.array([x.shape[0] for x in data])
-                self.data[key].append(lengths)
+                self.data[key][sub_key].extend([lengths])
                 break
 
     def split(
         self, number_of_test_examples: int, random_seed: int
     ) -> Tuple["RasaModelData", "RasaModelData"]:
-        """Create random hold out test set using stratified split."""
+        """Create random hold out test set using stratified split.
+
+        Args:
+            number_of_test_examples: Number of test examples.
+            random_seed: Random seed.
+
+        Returns:
+            A tuple of train and test RasaModelData.
+        """
 
         self._check_label_key()
 
-        if self.label_key is None:
+        if self.label_key is None or self.label_sub_key is None:
             # randomly split data as no label key is split
-            multi_values = [v for values in self.data.values() for v in values]
-            solo_values = [[] for values in self.data.values() for v in values]
+            multi_values = [
+                v
+                for attribute_data in self.data.values()
+                for data in attribute_data.values()
+                for v in data
+            ]
+            solo_values = [
+                []
+                for attribute_data in self.data.values()
+                for data in attribute_data.values()
+                for _ in data
+            ]
             stratify = None
         else:
             # make sure that examples for each label value are in both split sets
-            label_ids = self._create_label_ids(self.data[self.label_key][0])
+            label_ids = self._create_label_ids(
+                self.data[self.label_key][self.label_sub_key][0]
+            )
             label_counts = dict(zip(*np.unique(label_ids, return_counts=True, axis=0)))
 
             self._check_train_test_sizes(number_of_test_examples, label_counts)
@@ -182,11 +323,17 @@ class RasaModelData:
             # this operation can be performed only for labels
             # that contain several data points
             multi_values = [
-                v[counts > 1] for values in self.data.values() for v in values
+                f[counts > 1]
+                for attribute_data in self.data.values()
+                for features in attribute_data.values()
+                for f in features
             ]
             # collect data points that are unique for their label
             solo_values = [
-                v[counts == 1] for values in self.data.values() for v in values
+                f[counts == 1]
+                for attribute_data in self.data.values()
+                for features in attribute_data.values()
+                for f in features
             ]
 
             stratify = label_ids[counts > 1]
@@ -200,27 +347,43 @@ class RasaModelData:
 
         return self._convert_train_test_split(output_values, solo_values)
 
-    def get_signature(self) -> Dict[Text, List[FeatureSignature]]:
+    def get_signature(self) -> Dict[Text, Dict[Text, List[FeatureSignature]]]:
         """Get signature of RasaModelData.
 
         Signature stores the shape and whether features are sparse or not for every key.
+
+        Returns:
+            A dictionary of key and sub-key to a list of feature signatures
+            (same structure as the data attribute).
         """
 
         return {
-            key: [
-                FeatureSignature(
-                    True if isinstance(v[0], scipy.sparse.spmatrix) else False,
-                    v[0].shape,
-                )
-                for v in values
-            ]
-            for key, values in self.data.items()
+            key: {
+                sub_key: [
+                    FeatureSignature(
+                        True if isinstance(f[0], scipy.sparse.spmatrix) else False,
+                        f[0].shape[-1] if f[0].shape else None,
+                    )
+                    for f in features
+                ]
+                for sub_key, features in attribute_data.items()
+            }
+            for key, attribute_data in self.data.items()
         }
 
     def as_tf_dataset(
         self, batch_size: int, batch_strategy: Text = SEQUENCE, shuffle: bool = False
     ) -> tf.data.Dataset:
-        """Create tf dataset."""
+        """Create tf dataset.
+
+        Args:
+            batch_size: The batch size to use.
+            batch_strategy: The batch strategy to use.
+            shuffle: Boolean indicating whether the data should be shuffled or not.
+
+        Returns:
+            The tf.data.Dataset.
+        """
 
         shapes, types = self._get_shapes_types()
 
@@ -238,42 +401,59 @@ class RasaModelData:
         end: Optional[int] = None,
         tuple_sizes: Optional[Dict[Text, int]] = None,
     ) -> Tuple[Optional[np.ndarray]]:
-        """Slices model data into batch using given start and end value."""
+        """Slices model data into batch using given start and end value.
+
+        Args:
+            data: The data to prepare.
+            start: The start index of the batch
+            end: The end index of the batch
+            tuple_sizes: In case the feature is not present we propagate the batch with
+              None. Tuple sizes contains the number of how many None values to add for
+              what kind of feature.
+
+        Returns:
+            The features of the batch.
+        """
 
         if not data:
             data = self.data
 
         batch_data = []
 
-        for key, values in data.items():
-            # add None for not present values during processing
-            if not values:
-                if tuple_sizes:
-                    batch_data += [None] * tuple_sizes[key]
-                else:
-                    batch_data.append(None)
-                continue
+        for key, attribute_data in data.items():
+            for sub_key, f_data in attribute_data.items():
+                # add None for not present values during processing
+                if not f_data:
+                    if tuple_sizes:
+                        batch_data += [None] * tuple_sizes[key]
+                    else:
+                        batch_data.append(None)
+                    continue
 
-            for v in values:
-                if start is not None and end is not None:
-                    _data = v[start:end]
-                elif start is not None:
-                    _data = v[start:]
-                elif end is not None:
-                    _data = v[:end]
-                else:
-                    _data = v[:]
+                for v in f_data:
+                    if start is not None and end is not None:
+                        _data = v[start:end]
+                    elif start is not None:
+                        _data = v[start:]
+                    elif end is not None:
+                        _data = v[:end]
+                    else:
+                        _data = v[:]
 
-                if isinstance(_data[0], scipy.sparse.spmatrix):
-                    batch_data.extend(self._scipy_matrix_to_values(_data))
-                else:
-                    batch_data.append(self._pad_dense_data(_data))
+                    if isinstance(_data[0], scipy.sparse.spmatrix):
+                        batch_data.extend(self._scipy_matrix_to_values(_data))
+                    else:
+                        batch_data.append(self._pad_dense_data(_data))
 
         # len of batch_data is equal to the number of keys in model data
         return tuple(batch_data)
 
     def _get_shapes_types(self) -> Tuple:
-        """Extract shapes and types from model data."""
+        """Extract shapes and types from model data.
+
+        Returns:
+            A tuple of shapes and a tuple of types.
+        """
 
         types = []
         shapes = []
@@ -300,15 +480,23 @@ class RasaModelData:
             else:
                 types.append(tf.float32)
 
-        for values in self.data.values():
-            for v in values:
-                append_shape(v)
-                append_type(v)
+        for attribute_data in self.data.values():
+            for features in attribute_data.values():
+                for f in features:
+                    append_shape(f)
+                    append_type(f)
 
         return tuple(shapes), tuple(types)
 
     def _shuffled_data(self, data: Data) -> Data:
-        """Shuffle model data."""
+        """Shuffle model data.
+
+        Args:
+            data: The data to shuffle
+
+        Returns:
+            The shuffled data.
+        """
 
         ids = np.random.permutation(self.num_examples)
         return self._data_for_ids(data, ids)
@@ -319,15 +507,26 @@ class RasaModelData:
         This batching strategy puts rare classes approximately in every other batch,
         by repeating them. Mimics stratified batching, but also takes into account
         that more populated classes should appear more often.
-        """
 
+        Args:
+            data: The data.
+            batch_size: The batch size.
+            shuffle: Boolean indicating whether to shuffle the data or not.
+
+        Returns:
+            The balanced data.
+        """
         self._check_label_key()
 
         # skip balancing if labels are token based
-        if self.label_key is None or data[self.label_key][0][0].size > 1:
+        if (
+            self.label_key is None
+            or self.label_sub_key is None
+            or data[self.label_key][self.label_sub_key][0][0].size > 1
+        ):
             return data
 
-        label_ids = self._create_label_ids(data[self.label_key][0])
+        label_ids = self._create_label_ids(data[self.label_key][self.label_sub_key][0])
 
         unique_label_ids, counts_label_ids = np.unique(
             label_ids, return_counts=True, axis=0
@@ -345,7 +544,7 @@ class RasaModelData:
         # if a label was skipped in current batch
         skipped = [False] * num_label_ids
 
-        new_data = defaultdict(list)
+        new_data = defaultdict(lambda: defaultdict(list))
 
         while min(num_data_cycles) == 0:
             if shuffle:
@@ -357,20 +556,21 @@ class RasaModelData:
                 if num_data_cycles[index] > 0 and not skipped[index]:
                     skipped[index] = True
                     continue
-                else:
-                    skipped[index] = False
+
+                skipped[index] = False
 
                 index_batch_size = (
                     int(counts_label_ids[index] / self.num_examples * batch_size) + 1
                 )
 
-                for k, values in data_by_label[index].items():
-                    for i, v in enumerate(values):
-                        if len(new_data[k]) < i + 1:
-                            new_data[k].append([])
-                        new_data[k][i].append(
-                            v[data_idx[index] : data_idx[index] + index_batch_size]
-                        )
+                for key, attribute_data in data_by_label[index].items():
+                    for sub_key, features in attribute_data.items():
+                        for i, f in enumerate(features):
+                            if len(new_data[key][sub_key]) < i + 1:
+                                new_data[key][sub_key].append([])
+                            new_data[key][sub_key][i].append(
+                                f[data_idx[index] : data_idx[index] + index_batch_size]
+                            )
 
                 data_idx[index] += index_batch_size
                 if data_idx[index] >= counts_label_ids[index]:
@@ -380,17 +580,27 @@ class RasaModelData:
                 if min(num_data_cycles) > 0:
                     break
 
-        final_data = defaultdict(list)
-        for k, values in new_data.items():
-            for v in values:
-                final_data[k].append(np.concatenate(np.array(v)))
+        final_data = defaultdict(lambda: defaultdict(list))
+        for key, attribute_data in new_data.items():
+            for sub_key, features in attribute_data.items():
+                for f in features:
+                    final_data[key][sub_key].append(np.concatenate(np.array(f)))
 
         return final_data
 
     def _gen_batch(
         self, batch_size: int, batch_strategy: Text = SEQUENCE, shuffle: bool = False
     ) -> Generator[Tuple[Optional[np.ndarray]], None, None]:
-        """Generate batches."""
+        """Generate batches.
+
+        Args:
+            batch_size: The batch size
+            batch_strategy: The batch strategy.
+            shuffle: Boolean indicating whether to shuffle the data or not.
+
+        Returns:
+            A generator over the batches.
+        """
 
         data = self.data
         num_examples = self.num_examples
@@ -413,8 +623,16 @@ class RasaModelData:
 
     def _check_train_test_sizes(
         self, number_of_test_examples: int, label_counts: Dict[Any, int]
-    ):
-        """Check whether the test data set is too large or too small."""
+    ) -> None:
+        """Check whether the test data set is too large or too small.
+
+        Args:
+            number_of_test_examples: number of test examples
+            label_counts: number of labels
+
+        Raises:
+            A ValueError if the number of examples does not fit.
+        """
 
         if number_of_test_examples >= self.num_examples - len(label_counts):
             raise ValueError(
@@ -422,7 +640,7 @@ class RasaModelData:
                 f"train set should be at least equal to number of classes "
                 f"{len(label_counts)}."
             )
-        elif number_of_test_examples < len(label_counts):
+        if number_of_test_examples < len(label_counts):
             raise ValueError(
                 f"Test set of {number_of_test_examples} is too small. It should "
                 f"be at least equal to number of classes {label_counts}."
@@ -430,67 +648,114 @@ class RasaModelData:
 
     @staticmethod
     def _data_for_ids(data: Optional[Data], ids: np.ndarray) -> Data:
-        """Filter model data by ids."""
+        """Filter model data by ids.
 
-        new_data = defaultdict(list)
+        Args:
+            data: The data to filter
+            ids: The ids
+
+        Returns:
+            The filtered data
+        """
+
+        new_data = defaultdict(lambda: defaultdict(list))
 
         if data is None:
             return new_data
 
-        for k, values in data.items():
-            for v in values:
-                new_data[k].append(v[ids])
+        for key, attribute_data in data.items():
+            for sub_key, features in attribute_data.items():
+                for f in features:
+                    new_data[key][sub_key].append(f[ids])
         return new_data
 
     def _split_by_label_ids(
         self, data: Optional[Data], label_ids: np.ndarray, unique_label_ids: np.ndarray
     ) -> List["RasaModelData"]:
-        """Reorganize model data into a list of model data with the same labels."""
+        """Reorganize model data into a list of model data with the same labels.
+
+        Args:
+            data: The data
+            label_ids: The label ids
+            unique_label_ids: The unique label ids
+
+        Returns:
+            Reorganized RasaModelData
+        """
 
         label_data = []
         for label_id in unique_label_ids:
             matching_ids = label_ids == label_id
             label_data.append(
-                RasaModelData(self.label_key, self._data_for_ids(data, matching_ids))
+                RasaModelData(
+                    self.label_key,
+                    self.label_sub_key,
+                    self._data_for_ids(data, matching_ids),
+                )
             )
         return label_data
 
-    def _check_label_key(self):
-        if self.label_key is not None and (
-            self.label_key not in self.data or len(self.data[self.label_key]) > 1
+    def _check_label_key(self) -> None:
+        """Check if the label key exists.
+
+        Raises:
+            ValueError if the label key and sub-key is not in data.
+        """
+        if (
+            self.label_key is not None
+            and self.label_sub_key is not None
+            and (
+                self.label_key not in self.data
+                or self.label_sub_key not in self.data[self.label_key]
+                or len(self.data[self.label_key][self.label_sub_key]) > 1
+            )
         ):
-            raise ValueError(f"Key '{self.label_key}' not in RasaModelData.")
+            raise ValueError(
+                f"Key '{self.label_key}.{self.label_sub_key}' not in RasaModelData."
+            )
 
     def _convert_train_test_split(
         self, output_values: List[Any], solo_values: List[Any]
     ) -> Tuple["RasaModelData", "RasaModelData"]:
-        """Converts the output of sklearn's train_test_split into model data."""
+        """Converts the output of sklearn's train_test_split into model data.
 
-        data_train = defaultdict(list)
-        data_val = defaultdict(list)
+        Args:
+            output_values: output values of sklearn's train_test_split
+            solo_values: list of solo values
+
+        Returns:
+            The test and train RasaModelData
+        """
+
+        data_train = defaultdict(lambda: defaultdict(list))
+        data_val = defaultdict(lambda: defaultdict(list))
 
         # output_values = x_train, x_val, y_train, y_val, z_train, z_val, etc.
         # order is kept, e.g. same order as model data keys
 
         # train datasets have an even index
         index = 0
-        for key, values in self.data.items():
-            for _ in values:
-                data_train[key].append(
-                    self._combine_features(output_values[index * 2], solo_values[index])
-                )
-                index += 1
+        for key, attribute_data in self.data.items():
+            for sub_key, features in attribute_data.items():
+                for _ in features:
+                    data_train[key][sub_key].append(
+                        self._combine_features(
+                            output_values[index * 2], solo_values[index]
+                        )
+                    )
+                    index += 1
 
         # val datasets have an odd index
         index = 0
-        for key, values in self.data.items():
-            for _ in range(len(values)):
-                data_val[key].append(output_values[(index * 2) + 1])
-                index += 1
+        for key, attribute_data in self.data.items():
+            for sub_key, features in attribute_data.items():
+                for _ in features:
+                    data_val[key][sub_key].append(output_values[(index * 2) + 1])
+                    index += 1
 
         return (
-            RasaModelData(self.label_key, data_train),
-            RasaModelData(self.label_key, data_val),
+            RasaModelData(self.label_key, self.label_sub_key, data_train),
+            RasaModelData(self.label_key, self.label_sub_key, data_val),
         )
 
     @staticmethod
@@ -498,7 +763,15 @@ class RasaModelData:
         feature_1: Union[np.ndarray, scipy.sparse.spmatrix],
         feature_2: Union[np.ndarray, scipy.sparse.spmatrix],
     ) -> Union[np.ndarray, scipy.sparse.spmatrix]:
-        """Concatenate features."""
+        """Concatenate features.
+
+        Args:
+            feature_1: Features to concatenate.
+            feature_2: Features to concatenate.
+
+        Returns:
+            The combined features.
+        """
 
         if isinstance(feature_1, scipy.sparse.spmatrix) and isinstance(
             feature_2, scipy.sparse.spmatrix
@@ -518,6 +791,12 @@ class RasaModelData:
         For multi-label y, map each distinct row to a string representation
         using join because str(row) uses an ellipsis if len(row) > 1000.
         Idea taken from sklearn's stratify split.
+
+        Args:
+            label_ids: The label ids.
+
+        Returns:
+            The single dim label array.
         """
 
         if label_ids.ndim == 1:
@@ -539,6 +818,12 @@ class RasaModelData:
         """Pad data of different lengths.
 
         Sequential data is padded with zeros. Zeros are added to the end of data.
+
+        Args:
+            array_of_dense: The array to pad.
+
+        Returns:
+            The padded array.
         """
 
         if array_of_dense[0].ndim < 2:
@@ -559,7 +844,14 @@ class RasaModelData:
 
     @staticmethod
     def _scipy_matrix_to_values(array_of_sparse: np.ndarray) -> List[np.ndarray]:
-        """Convert a scipy matrix into indices, data, and shape."""
+        """Convert a scipy matrix into indices, data, and shape.
+
+        Args:
+            array_of_sparse: The sparse data array.
+
+        Returns:
+            A list of dense numpy arrays representing the sparse data.
+        """
 
         # we need to make sure that the matrices are coo_matrices otherwise the
         # transformation does not work (e.g. you cannot access x.row, x.col)
